import datetime
import threading

import jsonpickle

import websocket

from .algorum_types import *
from .concurrent_collections import *
from .remote_indicator_evaluator import *


class QuantEngineClient:
    MultiplyFactor = ((60 * 24) / 405)

    def __init__(self, url, apikey, launchmode, sid, user_id):
        self.Url = url
        self.ApiKey = apikey
        self.UserId = user_id
        self.LaunchMode = launchmode
        self.StrategyId = sid
        self.MessageHandlerMap = {}
        self.CorIdMessageMap = {}
        self.ws = None
        self.CorIdCounter = FastReadCounter()
        self.TickQueue = BoundedBlockingQueue(1000000)
        self.HandlerMessageQueue = BoundedBlockingQueue(1000000)
        self.stop_event = threading.Event()
        self.open_event = threading.Event()
        self.BacktestStartDate: datetime = datetime.utcnow()
        self.BacktestEndDate: datetime = datetime.utcnow()
        self.Evaluator = None
        self.LastProcessedTick = None
        self.ProgressPercent: float = 0.0
        self.Thread1 = None
        self.Thread2 = None
        self.Thread3 = None

        self.add_message_handler("tick", self.tick_handler)
        self.add_message_handler("order_update", self.order_update_handler)
        self.add_message_handler("stop", self.stop_handler)
        self.add_message_handler("error", self.error_handler)

        self.Evaluator: RemoteIndicatorEvaluator

        self.initialize()

    @staticmethod
    def get_date_format(timestamp: str):
        if timestamp.find("Z") > 0:
            return "%Y-%m-%dT%H:%M:%SZ"
        else:
            return "%Y-%m-%dT%H:%M:%S"

    def execute_async(self,
                      algorum_websocket_message: AlgorumWebsocketMessage) -> \
            AlgorumWebsocketMessage:
        async_waiter = AsyncWaiter()
        self.CorIdMessageMap[algorum_websocket_message.CorId] = async_waiter
        self.ws.send(jsonpickle.encode(algorum_websocket_message, False))
        async_waiter.WaiterEvent.wait()

        if async_waiter.Message.MessageType == AlgorumMessageType.ErrorResponse:
            raise AlgorumException(async_waiter.Message)

        return async_waiter.Message

    def send_async(self, algorum_websocket_message: AlgorumWebsocketMessage):
        self.ws.send(jsonpickle.encode(algorum_websocket_message, False))

    def publish_stats(self, stats):
        self.send_async(AlgorumWebsocketMessage(
            "publish_stats",
            AlgorumMessageType.Oneway,
            self.CorIdCounter.increment(),
            jsonpickle.encode(stats, False),
            None
        ))

    def set_data(self, key: str, value):
        quant_data = QuantData(key, jsonpickle.encode(value, False))
        response = self.execute_async(AlgorumWebsocketMessage(
            "set_data",
            AlgorumMessageType.Request,
            self.CorIdCounter.increment(),
            jsonpickle.encode(quant_data, False),
            None
        ))

        if response.MessageType == AlgorumMessageType.ErrorResponse:
            raise Exception(response.Error['ErrorMessage'])

    def get_data(self, key: str):
        request = AlgorumWebsocketMessage('get_data',
                                          AlgorumMessageType.Request,
                                          self.CorIdCounter.increment(),
                                          jsonpickle.encode(key, False), None)
        response = self.execute_async(request)

        if response.MessageType == AlgorumMessageType.ErrorResponse:
            raise Exception(response.Error['ErrorMessage'])

        value = jsonpickle.decode(response.JsonData)
        return value

    def add_message_handler(self, name, handler):
        self.MessageHandlerMap[name] = handler

    def error_handler(self, algorum_websocket_message):
        msg = jsonpickle.decode(algorum_websocket_message.JsonData)
        self.on_error(msg)

    def tick_handler(self, algorum_websocket_message):
        tick_data = jsonpickle.decode(algorum_websocket_message.JsonData)
        tick_data_obj = TickData(**tick_data)
        tick_data_obj.AlgorumMessage = algorum_websocket_message
        self.TickQueue.enqueue(tick_data_obj)

    def order_update_handler(self, algorum_websocket_message):
        order_dict = jsonpickle.decode(algorum_websocket_message.JsonData)
        order = Order(**order_dict)
        self.on_order_update(order)

        algorum_websocket_message.MessageType = AlgorumMessageType.Response
        algorum_websocket_message.JsonData = ""
        self.ws.send(jsonpickle.encode(algorum_websocket_message))

    def stop_handler(self, algorum_websocket_message):
        algorum_websocket_message.MessageType = AlgorumMessageType.Response
        algorum_websocket_message.JsonData = ""
        self.ws.send(jsonpickle.encode(algorum_websocket_message))
        self.stop_event.set()

    def wait(self):
        self.stop_event.wait()

    def subscribe_symbols(self, symbols):
        sub_symbols_request = \
            AlgorumWebsocketMessage('sub_symbols',
                                    AlgorumMessageType.Request,
                                    self.CorIdCounter.increment(),
                                    jsonpickle.encode(symbols, False), None)
        # print(sub_symbols_request)
        response = self.execute_async(sub_symbols_request)

        if response.MessageType == AlgorumMessageType.ErrorResponse:
            raise Exception(response.Error['ErrorMessage'])
        # print(response)

    def log(self, log_level: str, message: str):
        request = AlgorumWebsocketMessage('log',
                                          AlgorumMessageType.Oneway,
                                          self.CorIdCounter.increment(),
                                          jsonpickle.encode({
                                              'LogLevel': log_level,
                                              'Message': message
                                          }, False), None)
        self.send_async(request)

    def backtest(self, backtest_request: BacktestRequest):
        self.BacktestStartDate = backtest_request.StartDate
        self.BacktestEndDate = backtest_request.EndDate

        msg_request = \
            AlgorumWebsocketMessage('backtest',
                                    AlgorumMessageType.Request,
                                    self.CorIdCounter.increment(),
                                    jsonpickle.encode(backtest_request, False), None)
        # print(msg_request)
        response = self.execute_async(msg_request)

        if response.MessageType == AlgorumMessageType.ErrorResponse:
            raise Exception(response.Error['ErrorMessage'])

        # print(response)

    def start_trading(self, trading_request: TradingRequest):
        msg_request = \
            AlgorumWebsocketMessage('start_trading',
                                    AlgorumMessageType.Request,
                                    self.CorIdCounter.increment(),
                                    jsonpickle.encode(trading_request, False), None)
        # print(msg_request)
        response = self.execute_async(msg_request)

        if response.MessageType == AlgorumMessageType.ErrorResponse:
            raise Exception(response.Error['ErrorMessage'])

    def place_order(self, place_order_request: PlaceOrderRequest):
        request = AlgorumWebsocketMessage('place_order',
                                          AlgorumMessageType.Request,
                                          self.CorIdCounter.increment(),
                                          jsonpickle.encode(place_order_request, False), None)
        # print(request)
        response = self.execute_async(request)

        if response.MessageType == AlgorumMessageType.ErrorResponse:
            raise Exception(response.Error['ErrorMessage'])

        # print(response)
        order_id = jsonpickle.decode(response.JsonData)
        # print(order_id)
        return order_id

    def cancel_order(self, order_id: str) -> bool:
        request = AlgorumWebsocketMessage('cancel_order',
                                          AlgorumMessageType.Request,
                                          self.CorIdCounter.increment(),
                                          jsonpickle.encode(order_id, False), None)
        # print(request)
        response = self.execute_async(request)

        if response.MessageType == AlgorumMessageType.ErrorResponse:
            raise Exception(response.Error['ErrorMessage'])

        result: bool = jsonpickle.decode(response.JsonData)
        return result

    def create_indicator_evaluator(self, create_indicator_request: CreateIndicatorRequest) \
            -> RemoteIndicatorEvaluator:
        request = AlgorumWebsocketMessage('create_indicator_evaluator',
                                          AlgorumMessageType.Request,
                                          self.CorIdCounter.increment(),
                                          jsonpickle.encode(create_indicator_request, False), None)
        # print(request)
        response = self.execute_async(request)

        if response.MessageType == AlgorumMessageType.ErrorResponse:
            raise Exception(response.Error['ErrorMessage'])

        # print(response)

        uid = jsonpickle.decode(response.JsonData)
        # print(uid)

        self.Evaluator = RemoteIndicatorEvaluator(
            self,
            create_indicator_request.Symbol, uid)
        return self.Evaluator

    def get_stats(self, tick_data: TickData):
        raise NotImplementedError(self)

    def send_progress_async(self, tick_data: TickData):
        t_seconds = (self.BacktestEndDate - self.BacktestStartDate).total_seconds() * (70 / 100)
        processed_seconds = (datetime.strptime(tick_data.Timestamp,
                                               QuantEngineClient.get_date_format(tick_data.Timestamp)) -
                             datetime.strptime(self.LastProcessedTick.Timestamp,
                                               QuantEngineClient.get_date_format(self.LastProcessedTick.Timestamp))).total_seconds() * QuantEngineClient.MultiplyFactor

        progress_percent = (processed_seconds / t_seconds) * 100

        if (self.ProgressPercent + progress_percent >= 100) and not tick_data.LastTick:
            progress_percent = 0
        else:
            if tick_data.LastTick:
                self.ProgressPercent = 100

        if progress_percent >= 0.1 or progress_percent == 0 or tick_data.LastTick:
            if not tick_data.LastTick:
                self.ProgressPercent += progress_percent

            self.send_async(AlgorumWebsocketMessage(
                'publish_progress',
                AlgorumMessageType.Oneway,
                self.CorIdCounter.increment(),
                jsonpickle.encode(self.ProgressPercent, False), None))

            stats = self.get_stats(tick_data)
            self.publish_stats(stats)

            print('>>>>>>>>> Progress: ' + str(self.ProgressPercent))

            for k, v in stats.items():
                print('Key: ' + str(k) + ', Value: ' + str(v))

            self.LastProcessedTick = tick_data

        if self.ProgressPercent >= 100:
            self.log(LogLevel.Information, "100% Progress")
            self.stop_event.set()
            self.ws.close()

    def run(self):
        self.ws.run_forever()

    def on_error(self, msg: str):
        print(msg)

    def on_tick(self, tick_data: TickData):
        raise NotImplementedError(self)

    def on_order_update(self, order: Order):
        raise NotImplementedError(self)

    def process_tick(self):

        while 1:
            tick_data = self.TickQueue.dequeue()
            algorum_websocket_message = tick_data.AlgorumMessage
            tick_data.AlgorumMessage = None

            if tick_data is not None:
                last_processed_tick: TickData = self.LastProcessedTick

                if (self.LastProcessedTick is None) or \
                        (datetime.strptime(last_processed_tick.Timestamp,
                                           QuantEngineClient.get_date_format(last_processed_tick.Timestamp)).day !=
                         datetime.strptime(tick_data.Timestamp, QuantEngineClient.get_date_format(tick_data.Timestamp)).day):
                    self.LastProcessedTick = tick_data

                self.on_tick(tick_data)

                algorum_websocket_message.MessageType = AlgorumMessageType.Response
                algorum_websocket_message.JsonData = ""
                self.ws.send(jsonpickle.encode(algorum_websocket_message))
            else:
                break

    def handle_message(self):

        while 1:
            algorum_websocket_message = self.HandlerMessageQueue.dequeue()

            if algorum_websocket_message.Name in self.MessageHandlerMap:
                handler = self.MessageHandlerMap[algorum_websocket_message.Name]
            else:
                handler = None

            if handler is not None:
                handler(algorum_websocket_message)

    def initialize(self):
        def on_message(ws, message):
            # diagnostic message
            # print(message)

            msg_dict = jsonpickle.decode(message)
            msg = AlgorumWebsocketMessage(**msg_dict)

            if msg.Name in self.MessageHandlerMap:
                handler = self.MessageHandlerMap[msg.Name]
            else:
                handler = None

            if handler is None:
                if msg.MessageType == AlgorumMessageType.Request:
                    msg.MessageType = AlgorumMessageType.Response
                    msg.JsonData = ""
                    self.ws.send(jsonpickle.encode(msg))
                else:
                    if msg.MessageType == AlgorumMessageType.Response or \
                            msg.MessageType == AlgorumMessageType.ErrorResponse:
                        if msg.CorId in self.CorIdMessageMap:
                            async_waiter = self.CorIdMessageMap[msg.CorId]
                            async_waiter.Message = msg
                            del self.CorIdMessageMap[msg.CorId]
                            async_waiter.WaiterEvent.set()
            else:
                self.HandlerMessageQueue.enqueue(msg)

        def on_error(ws, error):
            print(error)

        def on_close(ws, error, msg):
            print('websockets closed')
            print(error)
            print(msg)
            self.stop_event.set()

        def on_open(ws):
            print('open')
            self.open_event.set()

        websocket.enableTrace(True)
        self.ws = websocket.WebSocketApp(self.Url,
                                         on_message=on_message,
                                         on_error=on_error,
                                         on_close=on_close,
                                         header={"x-algorum-userid": self.UserId})
        self.ws.on_open = on_open

        self.Thread1 = threading.Thread(target=self.run, daemon=True)
        self.Thread1.start()
        self.Thread2 = threading.Thread(target=self.process_tick, daemon=True)
        self.Thread2.start()
        self.Thread3 = threading.Thread(target=self.handle_message, daemon=True)
        self.Thread3.start()

        self.open_event.wait()



import threading
from collections import namedtuple
from datetime import datetime


class AsyncWaiter(object):
    def __init__(self):
        self.WaiterEvent = threading.Event()
        self.Message = None


class FastReadCounter(object):
    def __init__(self):
        self.value = 0
        self._lock = threading.Lock()

    def increment(self):
        with self._lock:
            self.value += 1
            return str(self.value)


class SymbolType:
    Stock = 'Stock'
    FuturesStock = 'FuturesStock'
    OptionsStock = 'OptionsStock'
    FuturesIndex = 'FuturesIndex'
    OptionsIndex = 'OptionsIndex'
    Index = 'Index'


class OptionType:
    Unspecified = 'None'
    PE = 'PE'
    CE = 'CE'
    XX = 'XX'


class SlippageType:
    BPS = 'BPS'
    TIME = 'TIME'


class LogLevel:
    Debug = 'Debug'
    Error = 'Error'
    Information = 'Information'
    Warning = 'Warning'


class FNOPeriodType:
    Monthly = 'Monthly'
    Weekly = 'Weekly'


class TradeSymbol(object):
    def __init__(self,
                 symboltype=None,
                 ticker=None,
                 fno_period_type=None,
                 fno_month=0,
                 fno_week=0,
                 option_type=None,
                 option_value=0,
                 script_token=0,
                 **kwargs):
        if symboltype is None:
            fill_obj(self, **kwargs)
        else:
            self.SymbolType = symboltype
            self.Ticker = ticker
            self.FNOPeriodType = fno_period_type
            self.FNOMonth = fno_month
            self.FNOWeek = fno_week
            self.OptionType = option_type
            self.OptionValue = option_value
            self.ScripToken = 0


def is_symbol_equal(source: TradeSymbol, target: TradeSymbol) -> bool:
    return source.Ticker == target.Ticker and \
           source.SymbolType == target.SymbolType and \
           source.OptionType == target.OptionType and \
           source.OptionValue == target.OptionValue and \
           source.FNOPeriodType == target.FNOPeriodType and \
           source.FNOMonth == target.FNOMonth and \
           source.FNOWeek == target.FNOWeek


class TickData(object):
    def __init__(self, symbol=None, date=None, timestamp=None, ltp=None, ltq=None,
                 bid=None, ask=None, last_tick=None, **kwargs):
        if symbol is None:
            fill_obj(self, **kwargs)
        else:
            self.Symbol = symbol
            self.Date = date
            self.Timestamp = timestamp
            self.LTP = ltp
            self.LTQ = ltq
            self.Bid = bid
            self.Ask = ask
            self.LastTick = last_tick


class AlgorumMessageType:
    Request = 'Request'
    Response = 'Response'
    Oneway = 'Oneway'
    ErrorResponse = 'ErrorResponse'


class AlgorumWebsocketError(object):
    def __init__(self):
        self.ErrorCode: int
        self.ErrorMessage: str
        self.ErrorStackTrace: str


class QuantData(object):
    def __init__(self, key=None, value=None, **kwargs):
        if key is None:
            fill_obj(self, **kwargs)
        else:
            self.Key = key
            self.Value = value


class AlgorumWebsocketMessage(object):
    def __init__(self, name=None, message_type=None, cor_id=None, json_data=None, error=None, **kwargs):
        if name is None:
            fill_obj(self, **kwargs)
        else:
            self.Name = name
            self.MessageType: AlgorumMessageType = message_type
            self.CorId = cor_id
            self.JsonData = json_data
            self.Error: AlgorumWebsocketError = error


class AlgorumException(BaseException):
    def __init__(self, msg: AlgorumWebsocketMessage):
        self.Message = msg


class BacktestRequest(object):
    def __init__(self,
                 start_date: datetime,
                 end_date: datetime,
                 uid: str,
                 apikey: str,
                 api_secret_key: str,
                 client_code: str,
                 password: str,
                 two_factor_auth: str,
                 sampling_time_in_seconds: int,
                 brokerage_platform: str,
                 capital: float):
        self.StartDate = start_date
        self.EndDate = end_date
        self.Uid = uid
        self.ApiKey = apikey
        self.ApiSecretKey = api_secret_key
        self.ClientCode = client_code
        self.Password = password
        self.TwoFactorAuth = two_factor_auth
        self.SamplingTimeInSeconds = sampling_time_in_seconds
        self.BrokeragePlatform = brokerage_platform
        self.Capital = capital


class TradingRequest(object):
    def __init__(self,
                 apikey: str,
                 api_secret_key: str,
                 client_code: str,
                 password: str,
                 two_factor_auth: str,
                 sampling_time_in_seconds: int,
                 brokerage_platform: str,
                 capital: float):
        self.ApiKey = apikey
        self.ApiSecretKey = api_secret_key
        self.ClientCode = client_code
        self.Password = password
        self.TwoFactorAuth = two_factor_auth
        self.SamplingTimeInSeconds = sampling_time_in_seconds
        self.BrokeragePlatform = brokerage_platform
        self.Capital = capital


class StrategyLaunchMode:
    Backtesting = 'Backtesting'
    PaperTrading = 'PaperTrading'
    LiveTrading = 'LiveTrading'


class CandlePeriod:
    Second = 'Second'
    Minute = 'Minute'
    Day = 'Day'


class CreateIndicatorRequest(object):
    def __init__(self, symbol: TradeSymbol, candle_period: CandlePeriod, period_span: int):
        self.Symbol = symbol
        self.CandlePeriod = candle_period
        self.PeriodSpan = period_span


class IndicatorRequest(object):
    def __init__(self, indicator: str, param_map):
        self.Indicator = indicator
        self.ParamMap = param_map


class GetIndicatorsRequest(object):
    def __init__(self, indicator_uid: str, indicator_requests):
        self.IndicatorUid = indicator_uid
        self.IndicatorRequests = indicator_requests


class PreloadCandlesRequest(object):
    def __init__(self, uid: str, candle_count: int, preload_end_time: datetime, api_key: str,
                 api_secret_key: str):
        self.IndicatorUid = uid
        self.CandleCount = candle_count
        self.PreloadEndTime = preload_end_time
        self.ApiKey = api_key
        self.ApiSecretKey = api_secret_key


class OrderProductType:
    Normal = 'Normal'
    Intraday = 'Intraday'
    CashAndCarry = 'CashAndCarry'


class OrderType:
    Market = 'Market'
    Limit = 'Limit'


class OrderDirection:
    Buy = 'Buy'
    Sell = 'Sell'


class TradeExchange:
    NSE = 'NSE'
    BSE = 'BSE'
    NFO = 'NFO'
    PAPER = 'PAPER'


class BrokeragePlatform:
    NorthEast = 'NorthEast'
    Alpaca = 'Alpaca'


class OrderStatus:
    Pending = 'Pending'
    Completed = 'Completed'
    Cancelled = 'Cancelled'
    Rejected = 'Rejected'


class PlaceOrderRequest(object):
    def __init__(self):
        self.Symbol = None
        self.Timestamp: datetime = datetime.utcnow()
        self.Validity = None
        self.TradeExchange = TradeExchange.PAPER
        self.OrderType = OrderType.Market
        self.OrderDirection = OrderDirection.Buy
        self.OrderProductType = OrderProductType.Normal
        self.Tag = None
        self.Quantity: float = 0.0
        self.Price: float = 0.0
        self.TriggerPrice: float = 0.0
        self.SlippageType = SlippageType.BPS
        self.Slippage = 0.0
        self.ValidTillTime: datetime = datetime.utcnow()


def fill_obj(obj, **kwargs):
    for a in kwargs:
        if type(kwargs[a]) is dict:
            obj_inner = namedtuple(a, kwargs[a].keys())(*kwargs[a].values())
            setattr(obj, a, obj_inner)
        else:
            setattr(obj, a, kwargs[a])


class Order(object):
    def __init__(self, **kwargs):
        self.TriggerPrice: float = 0.0
        self.OrderDirection = None
        self.Symbol: TradeSymbol = TradeSymbol()
        self.Tag = None
        self.StatusMessage = None
        self.Status = None
        self.Quantity: float = 0.0
        self.Product = None
        self.Price: float = 0.0
        self.PlacedBy = None
        self.PendingQuantity: float = 0.0
        self.ParentOrderId = None
        self.OrderType = None
        self.OrderTimestamp: datetime = datetime.utcnow()
        self.OrderId = None
        self.InstrumentToken: int = 0
        self.FilledQuantity: float = 0.0
        self.ExchangeTimestamp: datetime = datetime.utcnow()
        self.ExchangeOrderId = None
        self.Exchange = None
        self.DisclosedQuantity: float = 0.0
        self.CancelledQuantity: float = 0.0
        self.AveragePrice: float = 0.0
        self.Validity = None
        self.Variety = None
        self.BrokerageOrderJson = None
        self.LastTick: TickData = TickData()
        self.SlippageType = None
        self.Slippage = 0.0
        self.RunId = None
        self.ValidTillTime: datetime = datetime.utcnow()

        fill_obj(self, **kwargs)


class CrossAbove(object):
    def __init__(self):
        self.StopCrossOverUpdate = True
        self.CrossOverReached = False

    def evaluate(self, left_val: float, right_val: float) -> bool:
        if not self.StopCrossOverUpdate:
            self.CrossOverReached = left_val > right_val

            if self.CrossOverReached:
                self.StopCrossOverUpdate = True
        else:
            if left_val < right_val:
                self.StopCrossOverUpdate = False

            self.CrossOverReached = False

        return self.CrossOverReached


class CrossBelow(object):
    def __init__(self):
        self.StopCrossOverUpdate = True
        self.CrossOverReached = False

    def evaluate(self, left_val: float, right_val: float) -> bool:
        if not self.StopCrossOverUpdate:
            self.CrossOverReached = left_val < right_val

            if self.CrossOverReached:
                self.StopCrossOverUpdate = True
        else:
            if left_val > right_val:
                self.StopCrossOverUpdate = False

            self.CrossOverReached = False

        return self.CrossOverReached



from datetime import datetime

import jsonpickle

from .algorum_types import *


class RemoteIndicatorEvaluator(object):
    def __init__(self, client,
                 symbol: TradeSymbol,
                 uid: str):
        self.Uid = uid
        self.Symbol = symbol
        self.Client = client

    def clear_candles(self):
        response = self.Client.execute_async(AlgorumWebsocketMessage(
            'clear_indicator_candles',
            AlgorumMessageType.Request,
            self.Client.CorIdCounter.increment(),
            jsonpickle.encode(self.Uid, False), None))
        return

    def rsi(self, period: float):
        response = self.Client.execute_async(AlgorumWebsocketMessage(
            'get_indicators',
            AlgorumMessageType.Request,
            self.Client.CorIdCounter.increment(),
            jsonpickle.encode(
                GetIndicatorsRequest(
                    self.Uid,
                    [
                        IndicatorRequest(
                            'RSI',
                            {'period': period}
                        )
                    ]
                ), False), None))
        result_val = jsonpickle.decode(response.JsonData)
        return result_val[0]["Result"]

    def ema(self, period: float):
        response = self.Client.execute_async(AlgorumWebsocketMessage(
            'get_indicators',
            AlgorumMessageType.Request,
            self.Client.CorIdCounter.increment(),
            jsonpickle.encode(
                GetIndicatorsRequest(
                    self.Uid,
                    [
                        IndicatorRequest(
                            'EMA',
                            {'period': period}
                        )
                    ]
                ), False), None))
        result_val = jsonpickle.decode(response.JsonData)
        return result_val[0]["Result"]

    def abandoned_baby_bear(self, period: float):
        response = self.Client.execute_async(AlgorumWebsocketMessage(
            'get_indicators',
            AlgorumMessageType.Request,
            self.Client.CorIdCounter.increment(),
            jsonpickle.encode(
                GetIndicatorsRequest(
                    self.Uid,
                    [
                        IndicatorRequest(
                            'ABANDONEDBABYBEAR',
                            {'period': period}
                        )
                    ]
                ), False), None))
        result_val = jsonpickle.decode(response.JsonData)
        return result_val[0]["Result"]

    def abandoned_baby_bull(self, period: float):
        response = self.Client.execute_async(AlgorumWebsocketMessage(
            'get_indicators',
            AlgorumMessageType.Request,
            self.Client.CorIdCounter.increment(),
            jsonpickle.encode(
                GetIndicatorsRequest(
                    self.Uid,
                    [
                        IndicatorRequest(
                            'ABANDONEDBABYBULL',
                            {'period': period}
                        )
                    ]
                ), False), None))
        result_val = jsonpickle.decode(response.JsonData)
        return result_val[0]["Result"]

    def ad(self, period: float):
        response = self.Client.execute_async(AlgorumWebsocketMessage(
            'get_indicators',
            AlgorumMessageType.Request,
            self.Client.CorIdCounter.increment(),
            jsonpickle.encode(
                GetIndicatorsRequest(
                    self.Uid,
                    [
                        IndicatorRequest(
                            'AD',
                            {'period': period}
                        )
                    ]
                ), False), None))
        result_val = jsonpickle.decode(response.JsonData)
        return result_val[0]["Result"]

    def adosc(self, period: float):
        response = self.Client.execute_async(AlgorumWebsocketMessage(
            'get_indicators',
            AlgorumMessageType.Request,
            self.Client.CorIdCounter.increment(),
            jsonpickle.encode(
                GetIndicatorsRequest(
                    self.Uid,
                    [
                        IndicatorRequest(
                            'ADOSC',
                            {'period': period}
                        )
                    ]
                ), False), None))
        result_val = jsonpickle.decode(response.JsonData)
        return result_val[0]["Result"]

    def adx(self, period: float):
        response = self.Client.execute_async(AlgorumWebsocketMessage(
            'get_indicators',
            AlgorumMessageType.Request,
            self.Client.CorIdCounter.increment(),
            jsonpickle.encode(
                GetIndicatorsRequest(
                    self.Uid,
                    [
                        IndicatorRequest(
                            'ADX',
                            {'period': period}
                        )
                    ]
                ), False), None))
        result_val = jsonpickle.decode(response.JsonData)
        return result_val[0]["Result"]

    def adxr(self, period: float):
        response = self.Client.execute_async(AlgorumWebsocketMessage(
            'get_indicators',
            AlgorumMessageType.Request,
            self.Client.CorIdCounter.increment(),
            jsonpickle.encode(
                GetIndicatorsRequest(
                    self.Uid,
                    [
                        IndicatorRequest(
                            'ADXR',
                            {'period': period}
                        )
                    ]
                ), False), None))
        result_val = jsonpickle.decode(response.JsonData)
        return result_val[0]["Result"]

    def ao(self, period: float):
        response = self.Client.execute_async(AlgorumWebsocketMessage(
            'get_indicators',
            AlgorumMessageType.Request,
            self.Client.CorIdCounter.increment(),
            jsonpickle.encode(
                GetIndicatorsRequest(
                    self.Uid,
                    [
                        IndicatorRequest(
                            'AO',
                            {'period': period}
                        )
                    ]
                ), False), None))
        result_val = jsonpickle.decode(response.JsonData)
        return result_val[0]["Result"]

    def apo(self, short_period: float, long_period: float):
        response = self.Client.execute_async(AlgorumWebsocketMessage(
            'get_indicators',
            AlgorumMessageType.Request,
            self.Client.CorIdCounter.increment(),
            jsonpickle.encode(
                GetIndicatorsRequest(
                    self.Uid,
                    [
                        IndicatorRequest(
                            'APO',
                            {
                                'shortPeriod': short_period,
                                'longPeriod': long_period
                            }
                        )
                    ]
                ), False), None))
        result_val = jsonpickle.decode(response.JsonData)
        return result_val[0]["Result"]

    def aroon_down(self, period: float):
        response = self.Client.execute_async(AlgorumWebsocketMessage(
            'get_indicators',
            AlgorumMessageType.Request,
            self.Client.CorIdCounter.increment(),
            jsonpickle.encode(
                GetIndicatorsRequest(
                    self.Uid,
                    [
                        IndicatorRequest(
                            'AROONDOWN',
                            {'period': period}
                        )
                    ]
                ), False), None))
        result_val = jsonpickle.decode(response.JsonData)
        return result_val[0]["Result"]

    def aroon_osc(self, period: float):
        response = self.Client.execute_async(AlgorumWebsocketMessage(
            'get_indicators',
            AlgorumMessageType.Request,
            self.Client.CorIdCounter.increment(),
            jsonpickle.encode(
                GetIndicatorsRequest(
                    self.Uid,
                    [
                        IndicatorRequest(
                            'AROONOSC',
                            {'period': period}
                        )
                    ]
                ), False), None))
        result_val = jsonpickle.decode(response.JsonData)
        return result_val[0]["Result"]

    def aroon_up(self, period: float):
        response = self.Client.execute_async(AlgorumWebsocketMessage(
            'get_indicators',
            AlgorumMessageType.Request,
            self.Client.CorIdCounter.increment(),
            jsonpickle.encode(
                GetIndicatorsRequest(
                    self.Uid,
                    [
                        IndicatorRequest(
                            'AROONUP',
                            {'period': period}
                        )
                    ]
                ), False), None))
        result_val = jsonpickle.decode(response.JsonData)
        return result_val[0]["Result"]

    def atr(self, period: float):
        response = self.Client.execute_async(AlgorumWebsocketMessage(
            'get_indicators',
            AlgorumMessageType.Request,
            self.Client.CorIdCounter.increment(),
            jsonpickle.encode(
                GetIndicatorsRequest(
                    self.Uid,
                    [
                        IndicatorRequest(
                            'ATR',
                            {'period': period}
                        )
                    ]
                ), False), None))
        result_val = jsonpickle.decode(response.JsonData)
        return result_val[0]["Result"]

    def avg_price(self, period: float):
        response = self.Client.execute_async(AlgorumWebsocketMessage(
            'get_indicators',
            AlgorumMessageType.Request,
            self.Client.CorIdCounter.increment(),
            jsonpickle.encode(
                GetIndicatorsRequest(
                    self.Uid,
                    [
                        IndicatorRequest(
                            'AVGPRICE',
                            {'period': period}
                        )
                    ]
                ), False), None))
        result_val = jsonpickle.decode(response.JsonData)
        return result_val[0]["Result"]

    def bband_lower(self, period: float, stddev: float):
        response = self.Client.execute_async(AlgorumWebsocketMessage(
            'get_indicators',
            AlgorumMessageType.Request,
            self.Client.CorIdCounter.increment(),
            jsonpickle.encode(
                GetIndicatorsRequest(
                    self.Uid,
                    [
                        IndicatorRequest(
                            'BBANDLOWER',
                            {
                                'period': period,
                                'stddev': stddev
                            }
                        )
                    ]
                ), False), None))
        result_val = jsonpickle.decode(response.JsonData)
        return result_val[0]["Result"]

    def bband_mid(self, period: float, stddev: float):
        response = self.Client.execute_async(AlgorumWebsocketMessage(
            'get_indicators',
            AlgorumMessageType.Request,
            self.Client.CorIdCounter.increment(),
            jsonpickle.encode(
                GetIndicatorsRequest(
                    self.Uid,
                    [
                        IndicatorRequest(
                            'BBANDMID',
                            {
                                'period': period,
                                'stddev': stddev
                            }
                        )
                    ]
                ), False), None))
        result_val = jsonpickle.decode(response.JsonData)
        return result_val[0]["Result"]

    def bband_up(self, period: float, stddev: float):
        response = self.Client.execute_async(AlgorumWebsocketMessage(
            'get_indicators',
            AlgorumMessageType.Request,
            self.Client.CorIdCounter.increment(),
            jsonpickle.encode(
                GetIndicatorsRequest(
                    self.Uid,
                    [
                        IndicatorRequest(
                            'BBANDUP',
                            {
                                'period': period,
                                'stddev': stddev
                            }
                        )
                    ]
                ), False), None))
        result_val = jsonpickle.decode(response.JsonData)
        return result_val[0]["Result"]

    def big_black_candle(self, period: float):
        response = self.Client.execute_async(AlgorumWebsocketMessage(
            'get_indicators',
            AlgorumMessageType.Request,
            self.Client.CorIdCounter.increment(),
            jsonpickle.encode(
                GetIndicatorsRequest(
                    self.Uid,
                    [
                        IndicatorRequest(
                            'BIGBLACKCANDLE',
                            {'period': period}
                        )
                    ]
                ), False), None))
        result_val = jsonpickle.decode(response.JsonData)
        return result_val[0]["Result"]

    def black_marubozu(self, period: float):
        response = self.Client.execute_async(AlgorumWebsocketMessage(
            'get_indicators',
            AlgorumMessageType.Request,
            self.Client.CorIdCounter.increment(),
            jsonpickle.encode(
                GetIndicatorsRequest(
                    self.Uid,
                    [
                        IndicatorRequest(
                            'BLACKMARUBOZU',
                            {'period': period}
                        )
                    ]
                ), False), None))
        result_val = jsonpickle.decode(response.JsonData)
        return result_val[0]["Result"]

    def bop(self, period: float):
        response = self.Client.execute_async(AlgorumWebsocketMessage(
            'get_indicators',
            AlgorumMessageType.Request,
            self.Client.CorIdCounter.increment(),
            jsonpickle.encode(
                GetIndicatorsRequest(
                    self.Uid,
                    [
                        IndicatorRequest(
                            'BOP',
                            {'period': period}
                        )
                    ]
                ), False), None))
        result_val = jsonpickle.decode(response.JsonData)
        return result_val[0]["Result"]

    def candle_close(self, period: float):
        response = self.Client.execute_async(AlgorumWebsocketMessage(
            'get_indicators',
            AlgorumMessageType.Request,
            self.Client.CorIdCounter.increment(),
            jsonpickle.encode(
                GetIndicatorsRequest(
                    self.Uid,
                    [
                        IndicatorRequest(
                            'CANDLECLOSE',
                            {'period': period}
                        )
                    ]
                ), False), None))
        result_val = jsonpickle.decode(response.JsonData)
        return result_val[0]["Result"]

    def candle_high(self, period: float):
        response = self.Client.execute_async(AlgorumWebsocketMessage(
            'get_indicators',
            AlgorumMessageType.Request,
            self.Client.CorIdCounter.increment(),
            jsonpickle.encode(
                GetIndicatorsRequest(
                    self.Uid,
                    [
                        IndicatorRequest(
                            'CANDLEHIGH',
                            {'period': period}
                        )
                    ]
                ), False), None))
        result_val = jsonpickle.decode(response.JsonData)
        return result_val[0]["Result"]

    def candle_low(self, period: float):
        response = self.Client.execute_async(AlgorumWebsocketMessage(
            'get_indicators',
            AlgorumMessageType.Request,
            self.Client.CorIdCounter.increment(),
            jsonpickle.encode(
                GetIndicatorsRequest(
                    self.Uid,
                    [
                        IndicatorRequest(
                            'CANDLELOW',
                            {'period': period}
                        )
                    ]
                ), False), None))
        result_val = jsonpickle.decode(response.JsonData)
        return result_val[0]["Result"]

    def candle_open(self, period: float):
        response = self.Client.execute_async(AlgorumWebsocketMessage(
            'get_indicators',
            AlgorumMessageType.Request,
            self.Client.CorIdCounter.increment(),
            jsonpickle.encode(
                GetIndicatorsRequest(
                    self.Uid,
                    [
                        IndicatorRequest(
                            'CANDLEOPEN',
                            {'period': period}
                        )
                    ]
                ), False), None))
        result_val = jsonpickle.decode(response.JsonData)
        return result_val[0]["Result"]

    def cci(self, period: float):
        response = self.Client.execute_async(AlgorumWebsocketMessage(
            'get_indicators',
            AlgorumMessageType.Request,
            self.Client.CorIdCounter.increment(),
            jsonpickle.encode(
                GetIndicatorsRequest(
                    self.Uid,
                    [
                        IndicatorRequest(
                            'CCI',
                            {'period': period}
                        )
                    ]
                ), False), None))
        result_val = jsonpickle.decode(response.JsonData)
        return result_val[0]["Result"]

    def close(self):
        response = self.Client.execute_async(AlgorumWebsocketMessage(
            'get_indicators',
            AlgorumMessageType.Request,
            self.Client.CorIdCounter.increment(),
            jsonpickle.encode(
                GetIndicatorsRequest(
                    self.Uid,
                    [
                        IndicatorRequest('CLOSE', None)
                    ]
                ), False), None))
        result_val = jsonpickle.decode(response.JsonData)
        return result_val[0]["Result"]

    def cos(self, period: float):
        response = self.Client.execute_async(AlgorumWebsocketMessage(
            'get_indicators',
            AlgorumMessageType.Request,
            self.Client.CorIdCounter.increment(),
            jsonpickle.encode(
                GetIndicatorsRequest(
                    self.Uid,
                    [
                        IndicatorRequest(
                            'COS',
                            {'period': period}
                        )
                    ]
                ), False), None))
        result_val = jsonpickle.decode(response.JsonData)
        return result_val[0]["Result"]

    def dema(self, period: float):
        response = self.Client.execute_async(AlgorumWebsocketMessage(
            'get_indicators',
            AlgorumMessageType.Request,
            self.Client.CorIdCounter.increment(),
            jsonpickle.encode(
                GetIndicatorsRequest(
                    self.Uid,
                    [
                        IndicatorRequest(
                            'DEMA',
                            {'period': period}
                        )
                    ]
                ), False), None))
        result_val = jsonpickle.decode(response.JsonData)
        return result_val[0]["Result"]

    def doji(self, period: float):
        response = self.Client.execute_async(AlgorumWebsocketMessage(
            'get_indicators',
            AlgorumMessageType.Request,
            self.Client.CorIdCounter.increment(),
            jsonpickle.encode(
                GetIndicatorsRequest(
                    self.Uid,
                    [
                        IndicatorRequest(
                            'DOJI',
                            {'period': period}
                        )
                    ]
                ), False), None))
        result_val = jsonpickle.decode(response.JsonData)
        return result_val[0]["Result"]

    def dragonfly_doji(self, period: float):
        response = self.Client.execute_async(AlgorumWebsocketMessage(
            'get_indicators',
            AlgorumMessageType.Request,
            self.Client.CorIdCounter.increment(),
            jsonpickle.encode(
                GetIndicatorsRequest(
                    self.Uid,
                    [
                        IndicatorRequest(
                            'DRAGONFLYDOJI',
                            {'period': period}
                        )
                    ]
                ), False), None))
        result_val = jsonpickle.decode(response.JsonData)
        return result_val[0]["Result"]

    def ema(self, period: float):
        response = self.Client.execute_async(AlgorumWebsocketMessage(
            'get_indicators',
            AlgorumMessageType.Request,
            self.Client.CorIdCounter.increment(),
            jsonpickle.encode(
                GetIndicatorsRequest(
                    self.Uid,
                    [
                        IndicatorRequest(
                            'EMA',
                            {'period': period}
                        )
                    ]
                ), False), None))
        result_val = jsonpickle.decode(response.JsonData)
        return result_val[0]["Result"]

    def engulfing_bear(self, period: float):
        response = self.Client.execute_async(AlgorumWebsocketMessage(
            'get_indicators',
            AlgorumMessageType.Request,
            self.Client.CorIdCounter.increment(),
            jsonpickle.encode(
                GetIndicatorsRequest(
                    self.Uid,
                    [
                        IndicatorRequest(
                            'ENGULFINGBEAR',
                            {'period': period}
                        )
                    ]
                ), False), None))
        result_val = jsonpickle.decode(response.JsonData)
        return result_val[0]["Result"]

    def engulfing_bull(self, period: float):
        response = self.Client.execute_async(AlgorumWebsocketMessage(
            'get_indicators',
            AlgorumMessageType.Request,
            self.Client.CorIdCounter.increment(),
            jsonpickle.encode(
                GetIndicatorsRequest(
                    self.Uid,
                    [
                        IndicatorRequest(
                            'ENGULFINGBULL',
                            {'period': period}
                        )
                    ]
                ), False), None))
        result_val = jsonpickle.decode(response.JsonData)
        return result_val[0]["Result"]

    def evening_doji_star(self, period: float):
        response = self.Client.execute_async(AlgorumWebsocketMessage(
            'get_indicators',
            AlgorumMessageType.Request,
            self.Client.CorIdCounter.increment(),
            jsonpickle.encode(
                GetIndicatorsRequest(
                    self.Uid,
                    [
                        IndicatorRequest(
                            'EVENINGDOJISTART',
                            {'period': period}
                        )
                    ]
                ), False), None))
        result_val = jsonpickle.decode(response.JsonData)
        return result_val[0]["Result"]

    def evening_star(self, period: float):
        response = self.Client.execute_async(AlgorumWebsocketMessage(
            'get_indicators',
            AlgorumMessageType.Request,
            self.Client.CorIdCounter.increment(),
            jsonpickle.encode(
                GetIndicatorsRequest(
                    self.Uid,
                    [
                        IndicatorRequest(
                            'EVENINGSTAR',
                            {'period': period}
                        )
                    ]
                ), False), None))
        result_val = jsonpickle.decode(response.JsonData)
        return result_val[0]["Result"]

    def four_price_doji(self, period: float):
        response = self.Client.execute_async(AlgorumWebsocketMessage(
            'get_indicators',
            AlgorumMessageType.Request,
            self.Client.CorIdCounter.increment(),
            jsonpickle.encode(
                GetIndicatorsRequest(
                    self.Uid,
                    [
                        IndicatorRequest(
                            'FOURPRICEDOJI',
                            {'period': period}
                        )
                    ]
                ), False), None))
        result_val = jsonpickle.decode(response.JsonData)
        return result_val[0]["Result"]

    def grave_stone_doji(self, period: float):
        response = self.Client.execute_async(AlgorumWebsocketMessage(
            'get_indicators',
            AlgorumMessageType.Request,
            self.Client.CorIdCounter.increment(),
            jsonpickle.encode(
                GetIndicatorsRequest(
                    self.Uid,
                    [
                        IndicatorRequest(
                            'GRAVESTONEDOJI',
                            {'period': period}
                        )
                    ]
                ), False), None))
        result_val = jsonpickle.decode(response.JsonData)
        return result_val[0]["Result"]

    def hammer(self, period: float):
        response = self.Client.execute_async(AlgorumWebsocketMessage(
            'get_indicators',
            AlgorumMessageType.Request,
            self.Client.CorIdCounter.increment(),
            jsonpickle.encode(
                GetIndicatorsRequest(
                    self.Uid,
                    [
                        IndicatorRequest(
                            'HAMMER',
                            {'period': period}
                        )
                    ]
                ), False), None))
        result_val = jsonpickle.decode(response.JsonData)
        return result_val[0]["Result"]

    def hanging_man(self, period: float):
        response = self.Client.execute_async(AlgorumWebsocketMessage(
            'get_indicators',
            AlgorumMessageType.Request,
            self.Client.CorIdCounter.increment(),
            jsonpickle.encode(
                GetIndicatorsRequest(
                    self.Uid,
                    [
                        IndicatorRequest(
                            'HANGINGMAN',
                            {'period': period}
                        )
                    ]
                ), False), None))
        result_val = jsonpickle.decode(response.JsonData)
        return result_val[0]["Result"]

    def high(self):
        response = self.Client.execute_async(AlgorumWebsocketMessage(
            'get_indicators',
            AlgorumMessageType.Request,
            self.Client.CorIdCounter.increment(),
            jsonpickle.encode(
                GetIndicatorsRequest(
                    self.Uid,
                    [
                        IndicatorRequest('HIGH', None)
                    ]
                ), False), None))
        result_val = jsonpickle.decode(response.JsonData)
        return result_val[0]["Result"]

    def hma(self, period: float):
        response = self.Client.execute_async(AlgorumWebsocketMessage(
            'get_indicators',
            AlgorumMessageType.Request,
            self.Client.CorIdCounter.increment(),
            jsonpickle.encode(
                GetIndicatorsRequest(
                    self.Uid,
                    [
                        IndicatorRequest(
                            'HMA',
                            {'period': period}
                        )
                    ]
                ), False), None))
        result_val = jsonpickle.decode(response.JsonData)
        return result_val[0]["Result"]

    def inverted_hammer(self, period: float):
        response = self.Client.execute_async(AlgorumWebsocketMessage(
            'get_indicators',
            AlgorumMessageType.Request,
            self.Client.CorIdCounter.increment(),
            jsonpickle.encode(
                GetIndicatorsRequest(
                    self.Uid,
                    [
                        IndicatorRequest(
                            'INVERTEDHAMMER',
                            {'period': period}
                        )
                    ]
                ), False), None))
        result_val = jsonpickle.decode(response.JsonData)
        return result_val[0]["Result"]

    def kama(self, period: float):
        response = self.Client.execute_async(AlgorumWebsocketMessage(
            'get_indicators',
            AlgorumMessageType.Request,
            self.Client.CorIdCounter.increment(),
            jsonpickle.encode(
                GetIndicatorsRequest(
                    self.Uid,
                    [
                        IndicatorRequest(
                            'KAMA',
                            {'period': period}
                        )
                    ]
                ), False), None))
        result_val = jsonpickle.decode(response.JsonData)
        return result_val[0]["Result"]

    def long_legged_doji(self, period: float):
        response = self.Client.execute_async(AlgorumWebsocketMessage(
            'get_indicators',
            AlgorumMessageType.Request,
            self.Client.CorIdCounter.increment(),
            jsonpickle.encode(
                GetIndicatorsRequest(
                    self.Uid,
                    [
                        IndicatorRequest(
                            'LONGLEGGEDDOJI',
                            {'period': period}
                        )
                    ]
                ), False), None))
        result_val = jsonpickle.decode(response.JsonData)
        return result_val[0]["Result"]

    def low(self):
        response = self.Client.execute_async(AlgorumWebsocketMessage(
            'get_indicators',
            AlgorumMessageType.Request,
            self.Client.CorIdCounter.increment(),
            jsonpickle.encode(
                GetIndicatorsRequest(
                    self.Uid,
                    [
                        IndicatorRequest('LOW', None)
                    ]
                ), False), None))
        result_val = jsonpickle.decode(response.JsonData)
        return result_val[0]["Result"]

    def macd(self, short_period: float, long_period: float, signal_period: float):
        response = self.Client.execute_async(AlgorumWebsocketMessage(
            'get_indicators',
            AlgorumMessageType.Request,
            self.Client.CorIdCounter.increment(),
            jsonpickle.encode(
                GetIndicatorsRequest(
                    self.Uid,
                    [
                        IndicatorRequest(
                            'MACD',
                            {
                                'shortPeriod': short_period,
                                'longPeriod': long_period,
                                'signalPeriod': signal_period
                            }
                        )
                    ]
                ), False), None))
        result_val = jsonpickle.decode(response.JsonData)
        return result_val[0]["Result"]

    def macd_signal(self, short_period: float, long_period: float, signal_period: float):
        response = self.Client.execute_async(AlgorumWebsocketMessage(
            'get_indicators',
            AlgorumMessageType.Request,
            self.Client.CorIdCounter.increment(),
            jsonpickle.encode(
                GetIndicatorsRequest(
                    self.Uid,
                    [
                        IndicatorRequest(
                            'MACDSIGNAL',
                            {
                                'shortPeriod': short_period,
                                'longPeriod': long_period,
                                'signalPeriod': signal_period
                            }
                        )
                    ]
                ), False), None))
        result_val = jsonpickle.decode(response.JsonData)
        return result_val[0]["Result"]

    def marubozu(self, period: float):
        response = self.Client.execute_async(AlgorumWebsocketMessage(
            'get_indicators',
            AlgorumMessageType.Request,
            self.Client.CorIdCounter.increment(),
            jsonpickle.encode(
                GetIndicatorsRequest(
                    self.Uid,
                    [
                        IndicatorRequest(
                            'MARUBOZU',
                            {'period': period}
                        )
                    ]
                ), False), None))
        result_val = jsonpickle.decode(response.JsonData)
        return result_val[0]["Result"]

    def md(self, period: float):
        response = self.Client.execute_async(AlgorumWebsocketMessage(
            'get_indicators',
            AlgorumMessageType.Request,
            self.Client.CorIdCounter.increment(),
            jsonpickle.encode(
                GetIndicatorsRequest(
                    self.Uid,
                    [
                        IndicatorRequest(
                            'MD',
                            {'period': period}
                        )
                    ]
                ), False), None))
        result_val = jsonpickle.decode(response.JsonData)
        return result_val[0]["Result"]

    def mfi(self, period: float):
        response = self.Client.execute_async(AlgorumWebsocketMessage(
            'get_indicators',
            AlgorumMessageType.Request,
            self.Client.CorIdCounter.increment(),
            jsonpickle.encode(
                GetIndicatorsRequest(
                    self.Uid,
                    [
                        IndicatorRequest(
                            'MFI',
                            {'period': period}
                        )
                    ]
                ), False), None))
        result_val = jsonpickle.decode(response.JsonData)
        return result_val[0]["Result"]

    def minus_dmi(self, period: float):
        response = self.Client.execute_async(AlgorumWebsocketMessage(
            'get_indicators',
            AlgorumMessageType.Request,
            self.Client.CorIdCounter.increment(),
            jsonpickle.encode(
                GetIndicatorsRequest(
                    self.Uid,
                    [
                        IndicatorRequest(
                            'MINUSDMI',
                            {'period': period}
                        )
                    ]
                ), False), None))
        result_val = jsonpickle.decode(response.JsonData)
        return result_val[0]["Result"]

    def mom(self, period: float):
        response = self.Client.execute_async(AlgorumWebsocketMessage(
            'get_indicators',
            AlgorumMessageType.Request,
            self.Client.CorIdCounter.increment(),
            jsonpickle.encode(
                GetIndicatorsRequest(
                    self.Uid,
                    [
                        IndicatorRequest(
                            'MOM',
                            {'period': period}
                        )
                    ]
                ), False), None))
        result_val = jsonpickle.decode(response.JsonData)
        return result_val[0]["Result"]

    def morning_doji_star(self, period: float):
        response = self.Client.execute_async(AlgorumWebsocketMessage(
            'get_indicators',
            AlgorumMessageType.Request,
            self.Client.CorIdCounter.increment(),
            jsonpickle.encode(
                GetIndicatorsRequest(
                    self.Uid,
                    [
                        IndicatorRequest(
                            'MORNINGDOJISTAR',
                            {'period': period}
                        )
                    ]
                ), False), None))
        result_val = jsonpickle.decode(response.JsonData)
        return result_val[0]["Result"]

    def morning_star(self, period: float):
        response = self.Client.execute_async(AlgorumWebsocketMessage(
            'get_indicators',
            AlgorumMessageType.Request,
            self.Client.CorIdCounter.increment(),
            jsonpickle.encode(
                GetIndicatorsRequest(
                    self.Uid,
                    [
                        IndicatorRequest(
                            'MORNINGSTAR',
                            {'period': period}
                        )
                    ]
                ), False), None))
        result_val = jsonpickle.decode(response.JsonData)
        return result_val[0]["Result"]

    def natr(self, period: float):
        response = self.Client.execute_async(AlgorumWebsocketMessage(
            'get_indicators',
            AlgorumMessageType.Request,
            self.Client.CorIdCounter.increment(),
            jsonpickle.encode(
                GetIndicatorsRequest(
                    self.Uid,
                    [
                        IndicatorRequest(
                            'NATR',
                            {'period': period}
                        )
                    ]
                ), False), None))
        result_val = jsonpickle.decode(response.JsonData)
        return result_val[0]["Result"]

    def obv(self, period: float):
        response = self.Client.execute_async(AlgorumWebsocketMessage(
            'get_indicators',
            AlgorumMessageType.Request,
            self.Client.CorIdCounter.increment(),
            jsonpickle.encode(
                GetIndicatorsRequest(
                    self.Uid,
                    [
                        IndicatorRequest(
                            'OBV',
                            {'period': period}
                        )
                    ]
                ), False), None))
        result_val = jsonpickle.decode(response.JsonData)
        return result_val[0]["Result"]

    def open(self):
        response = self.Client.execute_async(AlgorumWebsocketMessage(
            'get_indicators',
            AlgorumMessageType.Request,
            self.Client.CorIdCounter.increment(),
            jsonpickle.encode(
                GetIndicatorsRequest(
                    self.Uid,
                    [
                        IndicatorRequest('OPEN', None)
                    ]
                ), False), None))
        result_val = jsonpickle.decode(response.JsonData)
        return result_val[0]["Result"]

    def plus_dmi(self, period: float):
        response = self.Client.execute_async(AlgorumWebsocketMessage(
            'get_indicators',
            AlgorumMessageType.Request,
            self.Client.CorIdCounter.increment(),
            jsonpickle.encode(
                GetIndicatorsRequest(
                    self.Uid,
                    [
                        IndicatorRequest(
                            'PLUSDMI',
                            {'period': period}
                        )
                    ]
                ), False), None))
        result_val = jsonpickle.decode(response.JsonData)
        return result_val[0]["Result"]

    def ppo(self, short_period: float, long_period: float):
        response = self.Client.execute_async(AlgorumWebsocketMessage(
            'get_indicators',
            AlgorumMessageType.Request,
            self.Client.CorIdCounter.increment(),
            jsonpickle.encode(
                GetIndicatorsRequest(
                    self.Uid,
                    [
                        IndicatorRequest(
                            'PPO',
                            {
                                'shortPeriod': short_period,
                                'longPeriod': long_period
                            }
                        )
                    ]
                ), False), None))
        result_val = jsonpickle.decode(response.JsonData)
        return result_val[0]["Result"]

    def prev_close(self):
        response = self.Client.execute_async(AlgorumWebsocketMessage(
            'get_indicators',
            AlgorumMessageType.Request,
            self.Client.CorIdCounter.increment(),
            jsonpickle.encode(
                GetIndicatorsRequest(
                    self.Uid,
                    [
                        IndicatorRequest('PREVCLOSE', None)
                    ]
                ), False), None))
        result_val = jsonpickle.decode(response.JsonData)
        return result_val[0]["Result"]

    def prev_high(self):
        response = self.Client.execute_async(AlgorumWebsocketMessage(
            'get_indicators',
            AlgorumMessageType.Request,
            self.Client.CorIdCounter.increment(),
            jsonpickle.encode(
                GetIndicatorsRequest(
                    self.Uid,
                    [
                        IndicatorRequest('PREVHIGH', None)
                    ]
                ), False), None))
        result_val = jsonpickle.decode(response.JsonData)
        return result_val[0]["Result"]

    def prev_low(self):
        response = self.Client.execute_async(AlgorumWebsocketMessage(
            'get_indicators',
            AlgorumMessageType.Request,
            self.Client.CorIdCounter.increment(),
            jsonpickle.encode(
                GetIndicatorsRequest(
                    self.Uid,
                    [
                        IndicatorRequest('PREVLOW', None)
                    ]
                ), False), None))
        result_val = jsonpickle.decode(response.JsonData)
        return result_val[0]["Result"]

    def prev_open(self):
        response = self.Client.execute_async(AlgorumWebsocketMessage(
            'get_indicators',
            AlgorumMessageType.Request,
            self.Client.CorIdCounter.increment(),
            jsonpickle.encode(
                GetIndicatorsRequest(
                    self.Uid,
                    [
                        IndicatorRequest('PREVOPEN', None)
                    ]
                ), False), None))
        result_val = jsonpickle.decode(response.JsonData)
        return result_val[0]["Result"]

    def psar(self, period: float, accl_factor_step: float, accl_factor_max: float):
        response = self.Client.execute_async(AlgorumWebsocketMessage(
            'get_indicators',
            AlgorumMessageType.Request,
            self.Client.CorIdCounter.increment(),
            jsonpickle.encode(
                GetIndicatorsRequest(
                    self.Uid,
                    [
                        IndicatorRequest(
                            'PSAR',
                            {
                                'period': period,
                                'acclFactorStep': accl_factor_step,
                                'acclFactorMax': accl_factor_max
                            }
                        )
                    ]
                ), False), None))
        result_val = jsonpickle.decode(response.JsonData)
        return result_val[0]["Result"]

    def roc(self, period: float):
        response = self.Client.execute_async(AlgorumWebsocketMessage(
            'get_indicators',
            AlgorumMessageType.Request,
            self.Client.CorIdCounter.increment(),
            jsonpickle.encode(
                GetIndicatorsRequest(
                    self.Uid,
                    [
                        IndicatorRequest(
                            'ROC',
                            {'period': period}
                        )
                    ]
                ), False), None))
        result_val = jsonpickle.decode(response.JsonData)
        return result_val[0]["Result"]

    def sin(self, period: float):
        response = self.Client.execute_async(AlgorumWebsocketMessage(
            'get_indicators',
            AlgorumMessageType.Request,
            self.Client.CorIdCounter.increment(),
            jsonpickle.encode(
                GetIndicatorsRequest(
                    self.Uid,
                    [
                        IndicatorRequest(
                            'SIN',
                            {'period': period}
                        )
                    ]
                ), False), None))
        result_val = jsonpickle.decode(response.JsonData)
        return result_val[0]["Result"]

    def sma(self, period: float):
        response = self.Client.execute_async(AlgorumWebsocketMessage(
            'get_indicators',
            AlgorumMessageType.Request,
            self.Client.CorIdCounter.increment(),
            jsonpickle.encode(
                GetIndicatorsRequest(
                    self.Uid,
                    [
                        IndicatorRequest(
                            'SMA',
                            {'period': period}
                        )
                    ]
                ), False), None))
        result_val = jsonpickle.decode(response.JsonData)
        return result_val[0]["Result"]

    def spinning_top(self, period: float):
        response = self.Client.execute_async(AlgorumWebsocketMessage(
            'get_indicators',
            AlgorumMessageType.Request,
            self.Client.CorIdCounter.increment(),
            jsonpickle.encode(
                GetIndicatorsRequest(
                    self.Uid,
                    [
                        IndicatorRequest(
                            'SPINNINGTOP',
                            {'period': period}
                        )
                    ]
                ), False), None))
        result_val = jsonpickle.decode(response.JsonData)
        return result_val[0]["Result"]

    def star(self, period: float):
        response = self.Client.execute_async(AlgorumWebsocketMessage(
            'get_indicators',
            AlgorumMessageType.Request,
            self.Client.CorIdCounter.increment(),
            jsonpickle.encode(
                GetIndicatorsRequest(
                    self.Uid,
                    [
                        IndicatorRequest(
                            'STAR',
                            {'period': period}
                        )
                    ]
                ), False), None))
        result_val = jsonpickle.decode(response.JsonData)
        return result_val[0]["Result"]

    def stddev(self, period: float):
        response = self.Client.execute_async(AlgorumWebsocketMessage(
            'get_indicators',
            AlgorumMessageType.Request,
            self.Client.CorIdCounter.increment(),
            jsonpickle.encode(
                GetIndicatorsRequest(
                    self.Uid,
                    [
                        IndicatorRequest(
                            'STDDEV',
                            {'period': period}
                        )
                    ]
                ), False), None))
        result_val = jsonpickle.decode(response.JsonData)
        return result_val[0]["Result"]

    def stoch_d(self, period: float, slowing_period: float, period_2: float):
        response = self.Client.execute_async(AlgorumWebsocketMessage(
            'get_indicators',
            AlgorumMessageType.Request,
            self.Client.CorIdCounter.increment(),
            jsonpickle.encode(
                GetIndicatorsRequest(
                    self.Uid,
                    [
                        IndicatorRequest(
                            'STOCHD',
                            {
                                'period': period,
                                'slowingPeriod': slowing_period,
                                'period2': period_2
                            }
                        )
                    ]
                ), False), None))
        result_val = jsonpickle.decode(response.JsonData)
        return result_val[0]["Result"]

    def stoch_k(self, period: float, slowing_period: float, period_2: float):
        response = self.Client.execute_async(AlgorumWebsocketMessage(
            'get_indicators',
            AlgorumMessageType.Request,
            self.Client.CorIdCounter.increment(),
            jsonpickle.encode(
                GetIndicatorsRequest(
                    self.Uid,
                    [
                        IndicatorRequest(
                            'STOCHK',
                            {
                                'period': period,
                                'slowingPeriod': slowing_period,
                                'period2': period_2
                            }
                        )
                    ]
                ), False), None))
        result_val = jsonpickle.decode(response.JsonData)
        return result_val[0]["Result"]

    def sub(self, period: float):
        response = self.Client.execute_async(AlgorumWebsocketMessage(
            'get_indicators',
            AlgorumMessageType.Request,
            self.Client.CorIdCounter.increment(),
            jsonpickle.encode(
                GetIndicatorsRequest(
                    self.Uid,
                    [
                        IndicatorRequest(
                            'SUB',
                            {'period': period}
                        )
                    ]
                ), False), None))
        result_val = jsonpickle.decode(response.JsonData)
        return result_val[0]["Result"]

    def sum(self, period: float):
        response = self.Client.execute_async(AlgorumWebsocketMessage(
            'get_indicators',
            AlgorumMessageType.Request,
            self.Client.CorIdCounter.increment(),
            jsonpickle.encode(
                GetIndicatorsRequest(
                    self.Uid,
                    [
                        IndicatorRequest(
                            'SUM',
                            {'period': period}
                        )
                    ]
                ), False), None))
        result_val = jsonpickle.decode(response.JsonData)
        return result_val[0]["Result"]

    def support_resistance(self, period: float, level: float, backtrack_candles: float):
        response = self.Client.execute_async(AlgorumWebsocketMessage(
            'get_indicators',
            AlgorumMessageType.Request,
            self.Client.CorIdCounter.increment(),
            jsonpickle.encode(
                GetIndicatorsRequest(
                    self.Uid,
                    [
                        IndicatorRequest(
                            'SUPPORTRESISTANCE',
                            {
                                'period': period,
                                'level': level,
                                'backtrackCandles': backtrack_candles
                            }
                        )
                    ]
                ), False), None))
        result_val = jsonpickle.decode(response.JsonData)
        return result_val[0]["ResultMap"]["support"], \
               result_val[0]["ResultMap"]["supportscore"], \
               result_val[0]["ResultMap"]["resistance"], \
               result_val[0]["ResultMap"]["resistancescore"]

    def tan(self, period: float):
        response = self.Client.execute_async(AlgorumWebsocketMessage(
            'get_indicators',
            AlgorumMessageType.Request,
            self.Client.CorIdCounter.increment(),
            jsonpickle.encode(
                GetIndicatorsRequest(
                    self.Uid,
                    [
                        IndicatorRequest(
                            'TAN',
                            {'period': period}
                        )
                    ]
                ), False), None))
        result_val = jsonpickle.decode(response.JsonData)
        return result_val[0]["Result"]

    def tema(self, period: float):
        response = self.Client.execute_async(AlgorumWebsocketMessage(
            'get_indicators',
            AlgorumMessageType.Request,
            self.Client.CorIdCounter.increment(),
            jsonpickle.encode(
                GetIndicatorsRequest(
                    self.Uid,
                    [
                        IndicatorRequest(
                            'TEMA',
                            {'period': period}
                        )
                    ]
                ), False), None))
        result_val = jsonpickle.decode(response.JsonData)
        return result_val[0]["Result"]

    def three_black_crows(self, period: float):
        response = self.Client.execute_async(AlgorumWebsocketMessage(
            'get_indicators',
            AlgorumMessageType.Request,
            self.Client.CorIdCounter.increment(),
            jsonpickle.encode(
                GetIndicatorsRequest(
                    self.Uid,
                    [
                        IndicatorRequest(
                            'THREEBLACKCROWS',
                            {'period': period}
                        )
                    ]
                ), False), None))
        result_val = jsonpickle.decode(response.JsonData)
        return result_val[0]["Result"]

    def three_white_soldiers(self, period: float):
        response = self.Client.execute_async(AlgorumWebsocketMessage(
            'get_indicators',
            AlgorumMessageType.Request,
            self.Client.CorIdCounter.increment(),
            jsonpickle.encode(
                GetIndicatorsRequest(
                    self.Uid,
                    [
                        IndicatorRequest(
                            'THREEWHITESOLDIERS',
                            {'period': period}
                        )
                    ]
                ), False), None))
        result_val = jsonpickle.decode(response.JsonData)
        return result_val[0]["Result"]

    def tra(self, period: float):
        response = self.Client.execute_async(AlgorumWebsocketMessage(
            'get_indicators',
            AlgorumMessageType.Request,
            self.Client.CorIdCounter.increment(),
            jsonpickle.encode(
                GetIndicatorsRequest(
                    self.Uid,
                    [
                        IndicatorRequest(
                            'TRA',
                            {'period': period}
                        )
                    ]
                ), False), None))
        result_val = jsonpickle.decode(response.JsonData)
        return result_val[0]["Result"]

    def trend(self, period: float):
        response = self.Client.execute_async(AlgorumWebsocketMessage(
            'get_indicators',
            AlgorumMessageType.Request,
            self.Client.CorIdCounter.increment(),
            jsonpickle.encode(
                GetIndicatorsRequest(
                    self.Uid,
                    [
                        IndicatorRequest(
                            'TREND',
                            {'period': period}
                        )
                    ]
                ), False), None))
        result_val = jsonpickle.decode(response.JsonData)
        return result_val[0]["Result"], result_val[0]["ResultMap"]["strength"]

    def trima(self, period: float):
        response = self.Client.execute_async(AlgorumWebsocketMessage(
            'get_indicators',
            AlgorumMessageType.Request,
            self.Client.CorIdCounter.increment(),
            jsonpickle.encode(
                GetIndicatorsRequest(
                    self.Uid,
                    [
                        IndicatorRequest(
                            'TRIMA',
                            {'period': period}
                        )
                    ]
                ), False), None))
        result_val = jsonpickle.decode(response.JsonData)
        return result_val[0]["Result"]

    def trix(self, period: float):
        response = self.Client.execute_async(AlgorumWebsocketMessage(
            'get_indicators',
            AlgorumMessageType.Request,
            self.Client.CorIdCounter.increment(),
            jsonpickle.encode(
                GetIndicatorsRequest(
                    self.Uid,
                    [
                        IndicatorRequest(
                            'TRIX',
                            {'period': period}
                        )
                    ]
                ), False), None))
        result_val = jsonpickle.decode(response.JsonData)
        return result_val[0]["Result"]

    def tsf(self, period: float):
        response = self.Client.execute_async(AlgorumWebsocketMessage(
            'get_indicators',
            AlgorumMessageType.Request,
            self.Client.CorIdCounter.increment(),
            jsonpickle.encode(
                GetIndicatorsRequest(
                    self.Uid,
                    [
                        IndicatorRequest(
                            'TSF',
                            {'period': period}
                        )
                    ]
                ), False), None))
        result_val = jsonpickle.decode(response.JsonData)
        return result_val[0]["Result"]

    def uosc(self, short_period: float, medium_period: float, long_period: float):
        response = self.Client.execute_async(AlgorumWebsocketMessage(
            'get_indicators',
            AlgorumMessageType.Request,
            self.Client.CorIdCounter.increment(),
            jsonpickle.encode(
                GetIndicatorsRequest(
                    self.Uid,
                    [
                        IndicatorRequest(
                            'UOSC',
                            {
                                'shortPeriod': short_period,
                                'mediumPeriod': medium_period,
                                'longPeriod': long_period
                            }
                        )
                    ]
                ), False), None))
        result_val = jsonpickle.decode(response.JsonData)
        return result_val[0]["Result"]

    def var(self, period: float):
        response = self.Client.execute_async(AlgorumWebsocketMessage(
            'get_indicators',
            AlgorumMessageType.Request,
            self.Client.CorIdCounter.increment(),
            jsonpickle.encode(
                GetIndicatorsRequest(
                    self.Uid,
                    [
                        IndicatorRequest(
                            'VAR',
                            {'period': period}
                        )
                    ]
                ), False), None))
        result_val = jsonpickle.decode(response.JsonData)
        return result_val[0]["Result"]

    def volatility(self, period: float):
        response = self.Client.execute_async(AlgorumWebsocketMessage(
            'get_indicators',
            AlgorumMessageType.Request,
            self.Client.CorIdCounter.increment(),
            jsonpickle.encode(
                GetIndicatorsRequest(
                    self.Uid,
                    [
                        IndicatorRequest(
                            'VOLATILITY',
                            {'period': period}
                        )
                    ]
                ), False), None))
        result_val = jsonpickle.decode(response.JsonData)
        return result_val[0]["Result"]

    def vwma(self, period: float):
        response = self.Client.execute_async(AlgorumWebsocketMessage(
            'get_indicators',
            AlgorumMessageType.Request,
            self.Client.CorIdCounter.increment(),
            jsonpickle.encode(
                GetIndicatorsRequest(
                    self.Uid,
                    [
                        IndicatorRequest(
                            'VWMA',
                            {'period': period}
                        )
                    ]
                ), False), None))
        result_val = jsonpickle.decode(response.JsonData)
        return result_val[0]["Result"]

    def white_marubozu(self, period: float):
        response = self.Client.execute_async(AlgorumWebsocketMessage(
            'get_indicators',
            AlgorumMessageType.Request,
            self.Client.CorIdCounter.increment(),
            jsonpickle.encode(
                GetIndicatorsRequest(
                    self.Uid,
                    [
                        IndicatorRequest(
                            'WHITEMARUBOZU',
                            {'period': period}
                        )
                    ]
                ), False), None))
        result_val = jsonpickle.decode(response.JsonData)
        return result_val[0]["Result"]

    def will_r(self, period: float):
        response = self.Client.execute_async(AlgorumWebsocketMessage(
            'get_indicators',
            AlgorumMessageType.Request,
            self.Client.CorIdCounter.increment(),
            jsonpickle.encode(
                GetIndicatorsRequest(
                    self.Uid,
                    [
                        IndicatorRequest(
                            'WILLR',
                            {'period': period}
                        )
                    ]
                ), False), None))
        result_val = jsonpickle.decode(response.JsonData)
        return result_val[0]["Result"]

    def wma(self, period: float):
        response = self.Client.execute_async(AlgorumWebsocketMessage(
            'get_indicators',
            AlgorumMessageType.Request,
            self.Client.CorIdCounter.increment(),
            jsonpickle.encode(
                GetIndicatorsRequest(
                    self.Uid,
                    [
                        IndicatorRequest(
                            'WMA',
                            {'period': period}
                        )
                    ]
                ), False), None))
        result_val = jsonpickle.decode(response.JsonData)
        return result_val[0]["Result"]

    def zlema(self, period: float):
        response = self.Client.execute_async(AlgorumWebsocketMessage(
            'get_indicators',
            AlgorumMessageType.Request,
            self.Client.CorIdCounter.increment(),
            jsonpickle.encode(
                GetIndicatorsRequest(
                    self.Uid,
                    [
                        IndicatorRequest(
                            'ZLEMA',
                            {'period': period}
                        )
                    ]
                ), False), None))
        result_val = jsonpickle.decode(response.JsonData)
        return result_val[0]["Result"]

    def preload_candles(self, candle_count: int, preload_end_time: datetime, api_key: str,
                        api_secret_key: str):
        response = self.Client.execute_async(AlgorumWebsocketMessage(
            'preload_candles',
            AlgorumMessageType.Request,
            self.Client.CorIdCounter.increment(),
            jsonpickle.encode(
                PreloadCandlesRequest(self.Uid, candle_count, preload_end_time, api_key, api_secret_key),
                False), None))
